/* HTTaP.c
   (C) 2011,2017,2020 Yann Guidon whygee@f-cpu.org
   The source code of this project is released under the AGPLv3 or later,
   see https://www.gnu.org/licenses/agpl-3.0.en.html

version 20170327: split from serv_11_CORS.c, a lot of renaming and translations
version 20170331: timeout and connection persistence work,
                  HTTaP and file serving appeared
version 20200415: removed IPv6, added more checks & some portability,
                  simplified some parts
version 20200416: solved a lot of little problems, works better than yesterday.
version 20200425: completely restructured around poll()
version 20200427: keepalive_counter removed
version 20200427: reworked...
version 20200501: missing close: finally solved

TODO : replace strncpy with memcpy

This file is included by HTTaP.h, where you can configure the parameters.
 */

#define HTTAP_TCP_BACKLOG 0 /* parameter for listen(2) */

#include <poll.h>

struct pollfd pollfd_array[2]={
  { SOCKET_ERROR, POLLIN, 0 }, // server socket, "readable" when accept() is possible
  { SOCKET_ERROR, POLLIN, 0 }  // Client socket, "writeable" only when open
};

/* persistent variables */
//int sock_server, sock_client; moved/aliased directly to the above array:
#define sock_server pollfd_array[0].fd
#define sock_client pollfd_array[1].fd

int //fcntl_flags, // it was used in a previous version, to ease the poll/block code
    nb_pollfd;  // =1 when waiting, =2 when client connection open

struct sockaddr_in sockaddr_client;
socklen_t size_client = sizeof(sockaddr_client);

#ifdef HTTAP_SERVE_FILES
struct stat stat_buf;
char *root_page;
uid_t prog_uid;
#endif

char header_buffer[HEADER_BUFFER_LEN],  // used by the MIME mechanism
      HTTaP_buffer[ HTTAP_BUFFER_LEN];  // Receive and emit buffer.

///////////////////////////////////////////////////////////////

int close_client() {
  if (sock_client != INVALID_SOCKET) {
    nb_pollfd=1;
    closesocket(sock_client);
    sock_client = INVALID_SOCKET;
    VERBOSE_PRINTF("\n###################################\n\n");
  }
//  else
//    VERBOSE_PRINTF(" !!! double close !!! \n");
  return 2; // force change of poll state
}

///////////////////////////////////////////////////////////////

void HTTaP_init() {
  int i;
  char *env_port, *b;
  struct addrinfo hints, chosen_ai;

#ifdef HTTAP_SERVE_FILES
  char *env_user, *env_group;
  struct passwd *pswd=NULL;
  struct group *grp=NULL;
  char *static_path;
#endif

/* Configuration */

  /* Port number */
  env_port = my_getenv(HTTAP_TCPPORT_NAME);
  if (!env_port) {
    env_port=HTTAP_TCPPORT;
    i = atol(env_port);
    if ((i < 2) || (i > 65535))
      show_exit("$" HTTAP_TCPPORT_NAME " must be between 1 and 65535\n", 0);
  }

  /* How long can a connection remain open ? */
  keepalive_seconds = HTTAP_KEEPALIVE;
  b = my_getenv(HTTAP_KEEPALIVE_NAME);
  if (b) {
    keepalive_seconds = atol(b);
    if ((keepalive_seconds < 2)
     || (keepalive_seconds > 200))
      show_exit("$" HTTAP_KEEPALIVE_NAME " must be between 2 and 200\n", 0);
  }
  VERBOSE_PRINTF("Port: %s, Keepalive: %ds\n", env_port, keepalive_seconds);

#ifdef HTTAP_SERVE_FILES
  /* Path to the directory that contains the static files */
  static_path = my_getenv(HTTAP_STATICPATH_NAME);
  if (!static_path)
    static_path = "."; /* The current working directory */

  /* Filename of the root URI (by default index.html) */
  root_page = my_getenv(HTTAP_ROOTPAGE_NAME);
  if (!root_page)
    root_page = HTTAP_ROOTPAGE;

  VERBOSE_PRINTF("Path: %s  Root page: %s\n", static_path, root_page);

  /* Check if we change the user and group */
  env_user = my_getenv(HTTAP_USER_NAME);
  if (env_user) {
    errno = 0;
    pswd = getpwnam(env_user);
    if (!pswd)
      show_exit("$" HTTAP_USER_NAME " is invalid ",0);

    VERBOSE_PRINTF("User: %s", env_user);
    prog_uid = pswd->pw_uid;

    env_group = my_getenv(HTTAP_GROUP_NAME);
    if (env_group) {
      errno = 0;
      grp = getgrnam(env_group);
      if (!grp)
        show_exit("$" HTTAP_GROUP_NAME "is invalid ", 0);

      VERBOSE_PRINTF(", Group: %s", env_group);
    }
  }
  else
    prog_uid = getuid();
  VERBOSE_PRINTF("\n");

  /* Check the path to access the static files */
  if (stat(static_path, &stat_buf)
   || !S_ISDIR(stat_buf.st_mode))
    show_exit("$" HTTAP_STATICPATH_NAME " : invalid path", 0);
  if (stat_buf.st_uid != prog_uid)
    show_exit("The program's user is not owner of $" HTTAP_STATICPATH_NAME "\n", 1);
  if (chdir(static_path))
    show_exit("Failed to change directory to $" HTTAP_STATICPATH_NAME , 1);
  if (chroot(static_path))
    perror("Warning: chroot() failure (are you root ?) ");

  /* Check the default root file */
  if (lstat(root_page, &stat_buf)
   || !S_ISREG(stat_buf.st_mode))
    show_exit("$" HTTAP_ROOTPAGE_NAME " : invalid file", 1);
  if (stat_buf.st_uid != prog_uid)
    show_exit("The user doesn't own $" HTTAP_ROOTPAGE_NAME "\n", 0);

  /* change the group THEN the user, after chroot() */
  if (pswd){ /* $HTTAP_USER exists */
    if (grp) {
      if (setgid(grp->gr_gid))
        show_exit("Failed to change running group ", 1);
    }
    if (setuid(pswd->pw_uid))
      show_exit("Failed to change running user ", 1);
  }
#endif

/*
   Open the socket
*/

  /* Describes the type of socket we want to open */
  memset(&hints, 0, sizeof(hints)); /* cleanup */
  hints.ai_family   = AF_INET;      /* IPv4 */
  hints.ai_socktype = SOCK_STREAM;  /* TCP */
  hints.ai_flags    = AI_PASSIVE;   /* server mode */

  if ((sock_server = create_IPv4_socket(NULL,
        env_port, &chosen_ai, &hints) ) == SOCKET_ERROR)
    show_exit("Failed to create the server.", 0);
  VERBOSE_PRINTF("\n");

#ifdef SO_REUSEADDR
  if (setsockopt(sock_server, SOL_SOCKET, SO_REUSEADDR, &const_one, sizeof(const_one)) == SOCKET_ERROR)
    net_perror("setsockopt(SO_REUSEADDR)");
#endif

#ifdef SO_REUSEPORT
  if (setsockopt(sock_server, SOL_SOCKET, SO_REUSEPORT, &const_one, sizeof(const_one)) == SOCKET_ERROR)
    net_perror("setsockopt(SO_REUSEPORT)");
#endif

  /* Connect the socket */
  if (bind(sock_server, chosen_ai.ai_addr, chosen_ai.ai_addrlen) == SOCKET_ERROR)
    net_exit("Failed to bind() the server");

  /* Listen on the socket */
  if (listen(sock_server, HTTAP_TCP_BACKLOG) == SOCKET_ERROR)
    net_exit("Failed to listen()");

//  /* Save the value to later unblock the socket */
//  if ((fcntl_flags = fcntl(sock_server, F_GETFL)) == SOCKET_ERROR)
//    net_exit("fcntl(F_GETFL) failed\n");

  nb_pollfd=1;
  VERBOSE_PRINTF("Server socket ready\n");
}


 /**************************************
       Examine the client's request
  **************************************/

int HTTaP_process_packet(int recv_len) {
  int ret=0, // this is the return value:
        // 0 is nothing happens
        // 1 if something happened and socket is to be kept alive.
        // 2 if socket has been close.
      len=sizeof(err400)-1; // size of the reply packet
  char *b=err400, // setup the default error.
       *fname;    // pointer to parse the URI
#ifdef HTTAP_SERVE_FILES
  int rfd, i, j, k, l, offset;
  char c, d, *Mime, *head;
#endif

#ifdef HTTAP_VERBOSE
  int n, max = HEADER_BUFFER_LEN-5;
  if (recv_len < HEADER_BUFFER_LEN-5)
    max = recv_len;

  printf("received %d bytes\n---------------------------\n", recv_len);
  k = 0;
  // safe-copy the first 2 lines of header/request to header_buffer,
  // max size is HEADER_BUFFER_LEN (give or take some terminator)
  for (n=0; (k<5) && (n<max); n++) {
    c=HTTaP_buffer[n];
    if ((c<' ') || (c >= 127)) {
      if ((c!=0xD) && (c!=0xA))
        c='!';
      else
        k++;
    }
    header_buffer[n]=c;
  }
  header_buffer[n+0]='\n';
  header_buffer[n+1]='*';
  header_buffer[n+2]='\n';
  header_buffer[n+3]=0;
  printf(header_buffer);
#endif


  // First, check the method :
  if ((recv_len >= GET_HEADER_LEN)
   && (memcmp(GET_header, HTTaP_buffer, GET_HEADER_LEN) == 0)) {
    VERBOSE_PRINTF("GET ! ");

    fname = &HTTaP_buffer[GET_HEADER_LEN]; // now point beyond the '/'
    recv_len -= GET_HEADER_LEN;

    /*
       HTTaP
     */

    if (( recv_len > 0) && (fname[0]=='?')) {

      fname++; // now point beyond the '?'
      VERBOSE_PRINTF("HTTaP : ");

      // check the Default root
      if ((recv_len == 1)
       ||((recv_len >  1)
            && ( (*fname==' ')
              || (*fname==0xD)
              || (*fname==0xA) ))) {
        VERBOSE_PRINTF("root\n");
        len=snprintf(HTTaP_buffer, HTTAP_BUFFER_LEN,
            "%sContent-Length: %lu\x0d\x0a\x0d\x0a%s",
            HTTaP_header, sizeof(HTTaP_json_root)-1, HTTaP_json_root);
        if (len >= HTTAP_BUFFER_LEN) {
          VERBOSE_PRINTF("Output buffer saturated ?\n");
          len=sizeof(err500)-1;
          b=err500;
        }
        else {
          b=HTTaP_buffer;
          ret=1;
        }
      } // root
      else {
        if ((recv_len >= PING_HEADER_LEN)
         && (memcmp(PING_header, fname, PING_HEADER_LEN)==0 )) {
               // The test is weak because *anything* is
               // allowed after the keyword to help prevent caching.
          VERBOSE_PRINTF("PING !\n");
          ret=1;
          len=sizeof(pong200)-1;
          b=pong200;
        }
        else {
              // insert your own HTTaP code here :-)

// redirect ?
// "HTTP/1.1 302 Found"
// "Location: "
              // result : error 400 unless b and len are modified.
              // don't forget to set these variables if the request ends well
              // b=HTTaP_buffer; // or somewhere else
              // len = ?
              // ret 1; ?
          VERBOSE_PRINTF(" key not found ");
          len=sizeof(err400key)-1;
          b=err400key;
        }
      }
      // now go to the send() section
    } // End of HTTaP domain

#ifdef HTTAP_SERVE_FILES
      /*
          The static file server
       */

    else {
      head=ok200; // This could be changed later to adjust/adapt with other packets

      if ((recv_len == 0)
       ||((recv_len >  0)
            && ( (*fname==' ')
              || (*fname==0xD)    // These ones were added
              || (*fname==0xA) ))) { // for manual tests with telnet
        VERBOSE_PRINTF("got root\n");
        // Substitute the blank URI with the index.html page.
        strncpy(fname, root_page, MAX_URI_LEN);
        recv_len=MAX_URI_LEN; // something is dangerous here ?
      }

      l=0;
      d=0;
      // scan the URI and clean it up
Scan_URI_Loop:
      c=fname[l];
      switch(c){
        case 0xD:
        case 0xA:
        case ' ':
        case 0:
          goto endscanURI;

        case '.':
          // prevent going to the root directory
          if (d=='.') goto send_short_block;
        case '/':
         // prevent going to the root directory
          if (d==0) goto send_short_block;
        case '-':  // only allowed characters
        case '_':  // (no '%' funky business)
          break;

        default:
          if (c< '0') goto send_short_block;
          if (c<='9') break;
          if (c< 'A') goto send_short_block;
          if (c<='Z') break;
          if((c< 'a')
          || (c> 'z')) goto send_short_block;
        // fold
      }
      d=c;
      if (++l > MAX_URI_LEN-(GET_HEADER_LEN+1)) {
        VERBOSE_PRINTF("URI size overflow: skip everything\n");
        goto send_short_block;
      }
      if (l < recv_len)
        goto Scan_URI_Loop; /* one more loop to process the next character */

endscanURI:
      fname[l]=0; /* coming from "case ' ':", terminate the string */
      VERBOSE_PRINTF("trying to read file: %s\n", fname);

      if (lstat(fname, &stat_buf) == -1) {
        perror("stat error : ");
        switch (errno) {
          case EACCES:
            VERBOSE_PRINTF("403: EACCES\n");
            len=sizeof(err403)-1;
            b=err404;
            break;

          case ENOTDIR:
          case ENOENT:
            VERBOSE_PRINTF("404: ENOTDIR/ENOENT\n");
            len=sizeof(err404)-1;
            b=err404;
            break;

          default:
            VERBOSE_PRINTF("500: others\n");
            len=sizeof(err500)-1;
            b=err500;
        }
        goto send_short_block;
      }

      if (!S_ISREG(stat_buf.st_mode)
      ||  stat_buf.st_uid != prog_uid) {
        VERBOSE_PRINTF("403: bad UID or file type ?\n");
        len=sizeof(err403)-1;
        b=err403;
        goto send_short_block;
      }

      /* locate the MIME TYPE file */
      Mime=""; // default : don't send the "Content-Type: " header string
      HTTaP_buffer[1]='.';
      HTTaP_buffer[2]='.'; // prepend "..." to the path
      HTTaP_buffer[3]='.';
      i=0; // type string length
      rfd=open(HTTaP_buffer+1, O_RDONLY);
      if (rfd == -1) {
        VERBOSE_PRINTF("Can't access the MIME type\n");
      }
      else {
        memcpy(&header_buffer, "Content-Type: ", 14);
        i=read(rfd, &(header_buffer[14]), HEADER_BUFFER_LEN-17);
        if (i == -1) {
          VERBOSE_PRINTF("Can't read the MIME type\n");
          i=0;
        }
        else {
          close(rfd);
          if (i>= HEADER_BUFFER_LEN-17) {
            printf("\nWARNING ! MIME type: file too long!\n");
            i=0;
          }
          else {
            Mime=header_buffer;
            Mime[i+14]=0xD;
            Mime[i+15]=0xA;
            Mime[i+16]=0; // zero termination
            VERBOSE_PRINTF("MIME type = %s\n", Mime);
          }
        }
      }

      /* open the actual data file now */
      rfd=open(HTTaP_buffer+5, O_RDONLY);
      if (rfd == -1) {
        perror("can't open data file ! ");
        len=sizeof(err500)-1;
        b=err500;
        goto send_short_block;
      }

      /* Assemble the header */
      offset=snprintf(HTTaP_buffer, HTTAP_BUFFER_LEN, head,
           Mime, (int)stat_buf.st_size);
      if (offset >= HTTAP_BUFFER_LEN) {
        VERBOSE_PRINTF("Header overflow, err500\n");
        len=sizeof(err500)-1;
        b=err500; // Check overflow, unlikely but you never know
        goto send_short_block;
      }

      VERBOSE_PRINTF("sending %d bytes header + %d bytes payload : ",
           offset, (int)stat_buf.st_size);

// TODO : sendfile(2) ?
// man 2 sendfile says : "not portable" :-/
// so here is the slow version:

      do {
        /* Read and appends the file's contents after the header */
        j=HTTAP_BUFFER_LEN-offset;
        i=read(rfd, HTTaP_buffer+offset, j);
        if (i == -1) {
          perror("can't read the data file ! ");
          if (offset) {  // the header can be changed
            len=sizeof(err500)-1;
            b=err500;
            goto send_short_block;
          }
          else
            goto close_sock_client; // simply abort
            // at this point because the header is already sent
        }
        k=i+offset;
        offset=0;

        if (k) { // don't call in the unlikely event that nothing has been read
          /* send the header and the available data */
          if (send(sock_client, HTTaP_buffer, k, 0) != k) {
            net_perror("Failed to send the reply");
            goto close_sock_client;
          }
          VERBOSE_PRINTF("%d ", k);
        }
      } while (i==j);

      VERBOSE_PRINTF("\n");
      return 1; // do not fold to sending the
                // small packet, return successfully !
    }
#endif  //  HTTAP_SERVE_FILES

  } // end of method GET
  else {
    if ((recv_len >= POST_HEADER_LEN)
     && (memcmp(POST_header, HTTaP_buffer, POST_HEADER_LEN) == 0)) {
      VERBOSE_PRINTF("POST !\n");

//
// POST stuff comes here.
//

    } // end of method POST
    else {
      VERBOSE_PRINTF("405: not GET or POST :-(\n");
      b=err405; // Method not allowed.
      len=sizeof(err405)-1;
    }
  }

  // Send a simple reply :
send_short_block:
  if (len >= HTTAP_BUFFER_LEN) // Should never happen...
    show_exit("Can't send this packet, too long !\n", 0);
  // Now b and len must be valid here.
  if (send(sock_client, b, len, 0) != len) {
    net_perror("Failed to send the reply");
    ret=0;
  }
  VERBOSE_PRINTF("Sent %d bytes of reply\n", len);
  if (ret==0) {
close_sock_client:
    return close_client();
  }

  return 1;
}


/////////////////////////////////////////////////////////////////////

int HTTaP_server(int timeout) {
   // return value :
   // 0 is nothing happens
   // 1 if something happened and socket is to be kept alive.
   // 2 if socket has been close.
  short ev;
  SOCKET_TYPE sock;
  int r,
    recv_len, // size of the received packet
    p=poll( &pollfd_array[0], nb_pollfd, timeout?0:H_poll_timeout);

  if (p==SOCKET_ERROR) {
    if (LastSockError() != INTR)
      net_exit("\npoll() error ");
    // now jump to the end of the function to check the timeout flag now.
  }
  else {  // did something happen ? check what in pollfd_array.

    // Check the client but only if it is open :
    if (nb_pollfd > 1) {
      ev = pollfd_array[1].revents;
      if (ev) {
        if (ev & ~POLLIN)
          net_exit("\npoll() problem on client socket \n");

        // finally, something to actually do, let's read that message (non-blocking way)
        recv_len=recv(sock_client, HTTaP_buffer, HTTAP_BUFFER_LEN, MSG_DONTWAIT);
        if (recv_len==SOCKET_ERROR) {
          r=LastSockError(x);
          if ((r != EWOULDBLOCK) && (r != EAGAIN)) {
            net_perror("Failed to read the client's request");
            return close_client();
          }
          VERBOSE_PRINTF(" recv() <= EWOULDBLOCK || EAGAIN ");
          // don't return yet, check the server socket, just in case ?
        }
        else {
          if (recv_len > 0)
            return HTTaP_process_packet(recv_len);

          // recv_len == 0:
          return close_client();  // socket closed by the peer, close the local socket
        }
      }
    } // End of processing client socket

    // Check if the server socket has a new request:
    ev = pollfd_array[0].revents;
    if (ev) {
      if (ev & ~POLLIN)
        net_exit("\npoll() problem on server socket \n");

      // ok : it's a connection request if you can read on that socket.
      if ((sock = accept(sock_server,
       (struct sockaddr *) &sockaddr_client, &size_client)) == SOCKET_ERROR)
        net_exit("accept(): Failure to wait for a network connection");

      // reject the new connection if we're already busy:
      if (sock_client != SOCKET_ERROR) {
        VERBOSE_PRINTF("Closing one spurious client\n");
        closesocket(sock);
        // don't return yet, check the timeout
      }
      else {
        // register our new client
        sock_client = sock;
        nb_pollfd=2;

        /* Disable Nagle : it's very good to have but not a critical feature */
        if (setsockopt(sock, IPPROTO_TCP, TCP_NODELAY, &const_one, sizeof(const_one)) < 0)
          net_perror("setsockopt(TCP_NODELAY) / Can't disable Nagle ");

#ifdef SHOW_CLIENT
        printf("* Connected to %s:%d\n",
             inet_ntoa(sockaddr_client.sin_addr), ntohs(sockaddr_client.sin_port));
#endif
        return 1; // yay ! something happened !
      } // new sock registered
    } // new socket accept() ?
  } // poll() had something new for us

  if (timeout)
    return close_client();

  return 0; // really nothing happened.
}
